@@ -8,25 +8,20 @@ like Racket's (arity-includes?).
 __all__ = ["getfunc",
            "arities", "arity_includes",
            "required_kwargs", "optional_kwargs", "kwargs",
-           "resolve_bindings", "tuplify_bindings",
+           "resolve_bindings", "resolve_bindings_partial", "tuplify_bindings",
            "UnknownArity"]
 
-from inspect import signature, Parameter, ismethod
 from collections import OrderedDict
+import copy
+from inspect import signature, Parameter, ismethod, BoundArguments, _empty
+import itertools
 import operator
 
-from .symbol import gensym
-
-try:  # Python 3.5+
-    from operator import matmul, imatmul
-except ImportError:  # pragma: no cover, Python 3.4 only.
-    NoSuchBuiltin = gensym("NoSuchBuiltin")
-    matmul = imatmul = NoSuchBuiltin
-
 class UnknownArity(ValueError):
     """Raised when the arity of a function cannot be inspected."""
 
 # HACK: some built-ins report incorrect arities (0, 0) at least in Python 3.4
+# TODO: re-test on 3.8, 3.9, 3.10, 3.11, 3.12 and on PyPy3 (3.8 and later), just to be sure.
 #
 # Full list of built-ins:
 #   https://docs.python.org/3/library/functions.html
@@ -127,7 +122,7 @@ _builtin_arities = {  # inspectable, but reporting incorrectly
                     operator.lshift: (2, 2),
                     operator.mod: (2, 2),
                     operator.mul: (2, 2),
-                    matmul: (2, 2),
+                    operator.matmul: (2, 2),
                     operator.neg: (1, 1),
                     operator.or_: (2, 2),
                     operator.pos: (1, 1),
@@ -154,7 +149,7 @@ _builtin_arities = {  # inspectable, but reporting incorrectly
                     operator.ilshift: (2, 2),
                     operator.imod: (2, 2),
                     operator.imul: (2, 2),
-                    imatmul: (2, 2),
+                    operator.imatmul: (2, 2),
                     operator.ior: (2, 2),
                     operator.ipow: (2, 2),
                     operator.irshift: (2, 2),
@@ -213,7 +208,7 @@ def arities(f):
     This uses inspect.signature; note that the signature of builtin functions
     cannot be inspected. This is worked around to some extent, but e.g.
     methods of built-in classes (such as ``list``) might not be inspectable
-    (at least on CPython < 3.7).
+    (at least on old CPython < 3.7).
 
     For bound methods, ``self`` or ``cls`` does not count toward the arity,
     because these are passed implicitly by Python. Note a `@classmethod` becomes
@@ -224,6 +219,9 @@ def arities(f):
      does not implicitly provide a `self`, because there is none to be had. This
      behavior is reflected in the return value of `arities`.)
 
+    If `f` is `@generic` (see `unpythonic.dispatch`), we scan its multimethods,
+    and return the smallest `min_arity` and the largest `max_arity`.
+
     Parameters:
         `f`: function
             The function to inspect.
@@ -245,6 +243,20 @@ def arities(f):
             return _builtin_arities[f]
     except TypeError:  # f is of an unhashable type
         pass
+
+    # Integration with the multiple-dispatch system (multimethods).
+    from .dispatch import isgeneric, list_methods  # circular import
+    if isgeneric(f):
+        min_lower = _infty
+        max_upper = 0
+        for (thecallable, type_signature) in list_methods(f):
+            lower, upper = arities(thecallable)  # let UnknownArity propagate
+            if lower < min_lower:
+                min_lower = lower
+            if upper > max_upper:
+                max_upper = upper
+        return min_lower, max_upper
+
     try:
         lower = 0
         upper = 0
@@ -265,25 +277,40 @@ def arities(f):
         raise UnknownArity(*e.args)
 
 def required_kwargs(f):
-    """Return a set containing the names of required name-only arguments of f.
+    """Return a set containing the names of required name-only arguments of `f`.
 
-    "Required": has no default.
+    *Required* means the parameter has no default.
 
-    Raises UnknownArity if inspection failed.
+    If `f` is `@generic` (see `unpythonic.dispatch`), we scan its multimethods,
+    and return the names of required kwargs accepted by *any* of its multimethods.
+
+    Raises `UnknownArity` if inspection failed.
     """
     return _kwargs(f, optionals=False)
 
 def optional_kwargs(f):
-    """Return a set containing the names of optional name-only arguments of f.
+    """Return a set containing the names of optional name-only arguments of `f`.
 
-    "Optional": has a default.
+    *Optional* means the parameter has a default.
 
-    Raises UnknownArity if inspection failed.
+    If `f` is `@generic` (see `unpythonic.dispatch`), we scan its multimethods,
+    and return the names of optional kwargs accepted by *any* of its multimethods.
+
+    Raises `UnknownArity` if inspection failed.
     """
     return _kwargs(f, optionals=True)
 
 def _kwargs(f, optionals=True):
     f, _ = getfunc(f)
+
+    # Integration with the multiple-dispatch system (multimethods).
+    from .dispatch import isgeneric, list_methods  # circular import
+    if isgeneric(f):
+        thekwargs = {}
+        for (thecallable, type_signature) in list_methods(f):
+            thekwargs.update(_kwargs(thecallable, optionals=optionals))
+        return thekwargs
+
     try:
         if optionals:
             pred = lambda v: v.default is not Parameter.empty  # optionals
@@ -312,14 +339,29 @@ def arity_includes(f, n):
     lower, upper = arities(f)
     return lower <= n <= upper
 
-# TODO: Can we replace this by `inspect.Signature.bind`, once we bump minimum Python 3.5+?
-# TODO: Python 3.4 has `inspect.callargs` (deprecated since 3.5).
+def resolve_bindings_partial(f, *args, **kwargs):
+    """Like `resolve_bindings`, but use `inspect.Signature.bind_partial`.
+
+    That is, it is acceptable for some parameters of `f` not to have a binding.
+    """
+    return _resolve_bindings(f, args, kwargs, _partial=True)
+
 def resolve_bindings(f, *args, **kwargs):
     """Resolve parameter bindings established by `f` when called with the given args and kwargs.
 
     This is an inspection tool, which does not actually call `f`. This is useful for memoizers
     and other similar decorators that need a canonical representation of `f`'s parameter bindings.
-    If you want a hashable result, postprocess the return value with `tuplify_bindings(result)`.
+
+    **NOTE**: This is a thin wrapper on top of `inspect.Signature.bind`, which was added in Python 3.5.
+    In `unpythonic` 0.14.2 and 0.14.3, we used to have our own implementation of the parameter binding
+    algorithm (that ran also on Python 3.4), but it is no longer needed, since as of v0.15.3,
+    we support only Python 3.8 and later.
+
+    The only thing we do beside call `inspect.Signature.bind` is that we apply default values
+    (from the definition of `f`) automatically.
+
+    The return value is an `inspect.BoundArguments`. If you want a hashable result,
+    postprocess the return value with `tuplify_bindings(result)`.
 
     For illustration, consider a simplistic memoizer::
 
@@ -366,159 +408,207 @@ def resolve_bindings(f, *args, **kwargs):
 
         f(42)
         f(a=42)  # now the cache hits
-
-    The return value of `resolve_bindings` is an `OrderedDict` with five keys:
-        args: `OrderedDict` of bindings made for regular parameters
-              (positional only, positional or keyword, keyword only).
-        vararg: `tuple` of arguments gathered by the vararg (`*args`) parameter
-                if the function definition has one; otherwise `None`.
-        vararg_name: `str`, the name of the vararg parameter; or `None`.
-        kwarg: `OrderedDict` of bindings gathered by `**kwargs` if the
-               function definition has one; otherwise `None`.
-        kwarg_name: `str`, the name of the kwarg parameter; or `None`.
-
-    **NOTE**:
-
-    We attempt to implement the exact same algorithm Python itself uses for
-    resolving argument bindings. The process is explained in the language
-    reference, although not in a step-by-step algorithmic form.
-
-        https://docs.python.org/3/reference/compound_stmts.html#function-definitions
-        https://docs.python.org/3/reference/expressions.html#calls
-
-    This function should report exactly those bindings that would actually be
-    established if `f` was actually called with the given `args` and `kwargs`.
-
-    If you encounter a case with any difference between what the result claims and
-    how Python itself assigns the bindings, that is a bug in our code. In such a
-    case, please report the issue, so it can be fixed, and then added to the unit
-    tests to ensure it won't come back.
     """
-    f, _ = getfunc(f)
-    params = signature(f).parameters
+    return _resolve_bindings(f, args, kwargs, _partial=False)
 
-    # https://docs.python.org/3/library/inspect.html#inspect.Signature
-    # https://docs.python.org/3/library/inspect.html#inspect.Parameter
-    poskinds = set((Parameter.POSITIONAL_ONLY,
-                    Parameter.POSITIONAL_OR_KEYWORD))
-    kwkinds = set((Parameter.POSITIONAL_OR_KEYWORD,
-                   Parameter.KEYWORD_ONLY))
-    varkinds = set((Parameter.VAR_POSITIONAL,
-                    Parameter.VAR_KEYWORD))
+def _resolve_bindings(f, args, kwargs, *, _partial):
+    thesignature = signature(f)
+    if _partial:
+        bound_arguments = thesignature.bind_partial(*args, **kwargs)
+    else:
+        bound_arguments = thesignature.bind(*args, **kwargs)
+    bound_arguments.apply_defaults()
+    return bound_arguments
 
-    index = {}
-    nposparams = 0
-    varpos = varkw = None
-    for slot, param in enumerate(params.values()):
-        if param.kind in poskinds:
-            nposparams += 1
-        if param.kind in kwkinds:
-            index[param.name] = slot
-        if param.kind == Parameter.VAR_POSITIONAL:
-            varpos = slot
-            varpos_name = param.name
-        elif param.kind == Parameter.VAR_KEYWORD:
-            varkw = slot
-            varkw_name = param.name
-
-    # https://docs.python.org/3/reference/compound_stmts.html#function-definitions
-    # https://docs.python.org/3/reference/expressions.html#calls
-    unassigned = object()  # gensym("unassigned"), but object() is much faster, and we don't need a label, or pickle support.
-    slots = [unassigned for _ in range(len(params))]  # yes, varparams too
-
-    # fill from positional arguments
-    for slot, (param, value) in enumerate(zip(params.values(), args)):
-        if param.kind in varkinds:  # these are always last in the function def
-            break
-        slots[slot] = value
-
-    if varpos is not None:
-        slots[varpos] = []
-    if varkw is not None:
-        slots[varkw] = OrderedDict()
-        vkdict = slots[varkw]
-
-    # gather excess positional arguments
-    if len(args) > nposparams:
-        if varpos is None:
-            raise TypeError("{}() takes {} positional arguments but {} were given".format(f.__name__,
-                                                                                          nposparams,
-                                                                                          len(args)))
-        slots[varpos] = args[nposparams:]
-
-    # fill from keyword arguments
-    for identifier, value in kwargs.items():
-        if identifier in index:
-            slot = index[identifier]
-            if slots[slot] is unassigned:
-                slots[slot] = value
-            else:
-                raise TypeError("{}() got multiple values for argument '{}'".format(f.__name__,
-                                                                                    identifier))
-        elif varkw is not None:  # gather excess keyword arguments
-            vkdict[identifier] = value
-        else:
-            raise TypeError("{}() got an unexpected keyword argument '{}'".format(f.__name__,
-                                                                                  identifier))
-
-    # fill missing with defaults from function definition
-    failures = []
-    for slot, param in enumerate(params.values()):
-        if slots[slot] is unassigned:
-            if param.default is Parameter.empty:
-                failures.append(param.name)
-            slots[slot] = param.default
-    # Python 3.6 goes so far to make this particular error message into proper
-    # English, that aping the standard error message takes the most effort here...
-    if failures:
-        if len(failures) == 1:
-            n1 = failures[0]
-            raise TypeError("{}() missing required positional argument: '{}'".format(f.__name__, n1))
-        if len(failures) == 2:
-            n1, n2 = failures
-            raise TypeError("{}() missing 2 required positional arguments: '{}' and '{}'".format(f.__name__, n1, n2))
-        wrapped = ["'{}'".format(x) for x in failures]
-        others = ", ".join(wrapped[:-1])
-        msg = "{}() missing {} required positional arguments: {}, and '{}'".format(f.__name__,
-                                                                                   len(failures),
-                                                                                   others,
-                                                                                   failures[-1])
-        raise TypeError(msg)
-
-    # build the result
-    regularargs = OrderedDict()
-    for param, value in zip(params.values(), slots):
-        if param.kind in varkinds:  # skip varpos, varkw
-            continue
-        regularargs[param.name] = value
-
-    # Naming of the fields matches `ast.arguments`
-    # https://greentreesnakes.readthedocs.io/en/latest/nodes.html#arguments
-    bindings = OrderedDict()
-    bindings["args"] = regularargs
-    bindings["vararg"] = slots[varpos] if varpos is not None else None
-    bindings["vararg_name"] = varpos_name if varpos is not None else None  # for introspection
-    bindings["kwarg"] = slots[varkw] if varkw is not None else None
-    bindings["kwarg_name"] = varkw_name if varkw is not None else None  # for introspection
-
-    return bindings
-
-def tuplify_bindings(bindings):
+def tuplify_bindings(bound_arguments):
     """Convert the return value of `resolve_bindings` into a hashable form.
 
     This is useful for memoizers and similar use cases, which need to use a
     representation of the bindings as a dictionary key.
 
-    The values stored in the `"args"` and `"kwarg"` keys, as well as `bindings`
-    itself, are converted from `OrderedDict` to `tuple` using `tuple(od.items())`.
-    The result is hashable, if all the arguments passed in the bindings are.
+    `bound_arguments` is an `inspect.BoundArguments` object.
+
+    In our return value, `bound_arguments.arguments` itself, as well as the value of
+    the `**kwargs` parameter contained in it, if any, are converted from `OrderedDict`
+    to `tuple` using `tuple(od.items())`.
+
+    The result is hashable, if all the passed arguments are.
+
+    See `resolve_bindings` for an example.
     """
-    def tuplify(od):
-        return tuple(od.items())
-    result = OrderedDict()
-    result["args"] = tuplify(bindings["args"])
-    result["vararg"] = bindings["vararg"]
-    result["vararg_name"] = bindings["vararg_name"]
-    result["kwarg"] = tuplify(bindings["kwarg"]) if bindings["kwarg"] is not None else None
-    result["kwarg_name"] = bindings["kwarg_name"]
-    return tuplify(result)
+    def tuplify(ordereddict):
+        return tuple(ordereddict.items())
+
+    # Tuplify the **kwargs dict.
+    #
+    # The information of which parameter it is, if any, is not contained in the
+    # `arguments` attribute of the `BoundArguments` instance; we need to scan
+    # the signature (stored in the `signature` attribute) against which the
+    # bindings were made.
+    for parameter in bound_arguments.signature.parameters.values():
+        if parameter.kind == Parameter.VAR_KEYWORD:
+            kwargs_param = parameter.name
+            break
+    else:
+        kwargs_param = None
+
+    if kwargs_param:
+        thearguments = copy.copy(bound_arguments.arguments)  # avoid mutating our input
+        thearguments[kwargs_param] = tuplify(thearguments[kwargs_param])
+    else:
+        thearguments = bound_arguments.arguments
+
+    return tuplify(thearguments)
+
+# This is `inspect.Signature.bind` from Python 3.8.5, modified for our purposes so we can determine
+# unbound *and extra* arguments (both positional and by-name) without raising a `TypeError`.
+# We need this for kwargs support in `curry`, because we want to pass through unmatched args and kwargs
+# (which otherwise trigger a `TypeError`).
+#
+# This is only for `curry`; all other code uses the standard implementation.
+#
+# Used under the PSF license. Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
+# 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Python Software Foundation; All Rights Reserved
+def _bind(thesignature, args, kwargs, *, partial):
+    """Private method. Don't use directly."""
+
+    arguments = OrderedDict()
+
+    parameters = iter(thesignature.parameters.values())
+    parameters_ex = ()
+    arg_vals = iter(args)
+
+    # These are added for `unpythonic`.
+    unbound_parameters = []
+    extra_args = []
+    extra_kwargs = OrderedDict()
+    kwargs = copy.copy(kwargs)  # the caller might need the original later
+
+    while True:
+        # Let's iterate through the positional arguments and corresponding
+        # parameters
+        try:
+            arg_val = next(arg_vals)
+        except StopIteration:
+            # No more positional arguments
+            try:
+                param = next(parameters)
+            except StopIteration:
+                # No more parameters. That's it. Just need to check that
+                # we have no `kwargs` after this while loop
+                break
+            else:
+                if param.kind == Parameter.VAR_POSITIONAL:
+                    # That's OK, just empty *args.  Let's start parsing
+                    # kwargs
+                    break
+                elif param.name in kwargs:
+                    if param.kind == Parameter.POSITIONAL_ONLY:
+                        msg = '{arg!r} parameter is positional only, ' \
+                              'but was passed as a keyword'
+                        msg = msg.format(arg=param.name)
+                        raise TypeError(msg) from None
+                    parameters_ex = (param,)
+                    break
+                elif (param.kind == Parameter.VAR_KEYWORD or
+                                            param.default is not _empty):
+                    # That's fine too - we have a default value for this
+                    # parameter.  So, lets start parsing `kwargs`, starting
+                    # with the current parameter
+                    parameters_ex = (param,)
+                    break
+                else:
+                    # No default, not VAR_KEYWORD, not VAR_POSITIONAL,
+                    # not in `kwargs`
+                    if partial:
+                        parameters_ex = (param,)
+                        break
+                    else:
+                        # msg = 'missing a required argument: {arg!r}'
+                        # msg = msg.format(arg=param.name)
+                        # raise TypeError(msg) from None
+                        unbound_parameters.append(param)
+        else:
+            # We have a positional argument to process
+            try:
+                param = next(parameters)
+            except StopIteration:
+                # raise TypeError('too many positional arguments') from None
+                extra_args.append(arg_val)
+            else:
+                if param.kind in (Parameter.VAR_KEYWORD, Parameter.KEYWORD_ONLY):
+                    # Looks like we have no parameter for this positional
+                    # argument
+                    # raise TypeError(
+                    #     'too many positional arguments') from None
+                    extra_args.append(arg_val)
+
+                if param.kind == Parameter.VAR_POSITIONAL:
+                    # We have an '*args'-like argument, let's fill it with
+                    # all positional arguments we have left and move on to
+                    # the next phase
+                    values = [arg_val]
+                    values.extend(arg_vals)
+                    arguments[param.name] = tuple(values)
+                    break
+
+                if param.name in kwargs and param.kind != Parameter.POSITIONAL_ONLY:
+                    raise TypeError(
+                        'multiple values for argument {arg!r}'.format(
+                            arg=param.name)) from None
+
+                arguments[param.name] = arg_val
+
+    # Now, we iterate through the remaining parameters to process
+    # keyword arguments
+    kwargs_param = None
+    for param in itertools.chain(parameters_ex, parameters):
+        if param.kind == Parameter.VAR_KEYWORD:
+            # Memorize that we have a '**kwargs'-like parameter
+            kwargs_param = param
+            continue
+
+        if param.kind == Parameter.VAR_POSITIONAL:
+            # Named arguments don't refer to '*args'-like parameters.
+            # We only arrive here if the positional arguments ended
+            # before reaching the last parameter before *args.
+            continue
+
+        param_name = param.name
+        try:
+            arg_val = kwargs.pop(param_name)
+        except KeyError:
+            # We have no value for this parameter.  It's fine though,
+            # if it has a default value, or it is an '*args'-like
+            # parameter, left alone by the processing of positional
+            # arguments.
+            if (not partial and param.kind != Parameter.VAR_POSITIONAL and
+                                                param.default is _empty):
+                # raise TypeError('missing a required argument: {arg!r}'.
+                #                 format(arg=param_name)) from None
+                unbound_parameters.append(param)
+
+        else:
+            if param.kind == Parameter.POSITIONAL_ONLY:
+                # This should never happen in case of a properly built
+                # Signature object (but let's have this check here
+                # to ensure correct behaviour just in case)
+                raise TypeError('{arg!r} parameter is positional only, '
+                                'but was passed as a keyword'.
+                                format(arg=param.name))
+
+            arguments[param_name] = arg_val
+
+    if kwargs:
+        if kwargs_param is not None:
+            # Process our '**kwargs'-like parameter
+            arguments[kwargs_param.name] = kwargs
+        else:
+            # raise TypeError(
+            #     'got an unexpected keyword argument {arg!r}'.format(
+            #         arg=next(iter(kwargs))))
+            extra_kwargs.update(kwargs)
+
+    return (BoundArguments(thesignature, arguments),
+            tuple(unbound_parameters),
+            (tuple(extra_args), extra_kwargs))
