@@ -57,7 +57,8 @@ __all__ = ["signal", "error",
            "available_restarts", "available_handlers",
            "restarts", "with_restarts",
            "handlers",
-           "ControlError"]
+           "ControlError",
+           "resignal_in", "resignal"]
 
 import threading
 from collections import deque, namedtuple
@@ -68,7 +69,8 @@ import warnings
 
 from .collections import box, unbox
 from .arity import arity_includes, UnknownArity
-from .misc import namelambda, equip_with_traceback, safeissubclass
+from .excutil import equip_with_traceback
+from .misc import namelambda, safeissubclass
 
 _stacks = threading.local()
 def _ensure_stacks():  # per-thread init
@@ -86,7 +88,7 @@ class ControlError(Exception):
     when no handler handles the signal.
     """
 
-def signal(condition, *, cause=None):
+def signal(condition, *, cause=None, protocol=None):
     """Signal a condition.
 
     Signaling a condition works similarly to raising an exception (pass an
@@ -113,7 +115,18 @@ def signal(condition, *, cause=None):
     perform the restart and continue.
 
     If none of the matching handlers invokes a restart, `signal` returns
-    normally. There is no meaningful return value, it is always `None`.
+    normally.
+
+    For most use cases, the return value is not needed. But for defining
+    custom error-handling protocols on top of `signal`, it can be very useful.
+
+    The return value is the input `condition`, canonized to an instance
+    (even if originally, an exception *type* was passed to `signal`),
+    with its `__cause__` and `__protocol__` attributes filled in,
+    and with a traceback attached. For example, the `error` protocol
+    uses the return value to chain the unhandled signal properly into
+    a `ControlError` exception; as a result, the error report looks
+    like a standard exception chain, with nice-looking tracebacks.
 
     If you want to error out on unhandled conditions, see `error`, which is
     otherwise the same as `signal`, except it raises if `signal` would have
@@ -124,6 +137,16 @@ def signal(condition, *, cause=None):
     keyword, it essentially performs a `signal ... from ...`. The default
     `cause=None` performs a plain `signal ...`.
 
+    The optional `protocol` argument is a low-level detail, meant for use by
+    error-handling protocols (including custom ones).
+
+    The `protocol` is stored into the `__protocol__` attribute of the condition
+    instance. It is the callable that was used to perform the signaling. If not
+    given, it defaults to the `signal` function itself. The main use case is for
+    `resignal` (for signal type conversion); using this information, it can
+    automatically emit the new signal using the same protocol that was used
+    for the original signal (so that e.g. an `error` remains an `error`).
+
     **Notes**
 
     This condition system is implemented on top of exceptions. The magic trick
@@ -139,9 +162,8 @@ def signal(condition, *, cause=None):
     You can signal any exception or warning object, both builtins and any
     custom ones.
 
-    On Python 3.7 and later, the exception object representing the signaled
-    condition is equipped with a traceback, just like a raised exception.
-    On Python 3.6 this is not possible, so the traceback is `None`.
+    The exception object representing the signaled condition is equipped
+    with a traceback, just like a raised exception.
     """
     # Since the handler is called normally, we don't unwind the call stack,
     # remaining inside the `signal()` call in the low-level code.
@@ -149,6 +171,16 @@ def signal(condition, *, cause=None):
     # The unwinding, when it occurs, is performed when `invoke` is
     # called from inside the condition handler in the user code.
 
+    # stacklevel: in the traceback, omit equip_with_traceback(), _prepare_signal_instance(), and signal().
+    #
+    # We can't have signal() there, because it would look like the call to `_prepare_signal_instance`
+    # was the cause of the signal, which is nonsense.
+    #
+    # Nicely, the resulting stack trace happens to be similar to how Python handles `raise` - the use site
+    # of `raise` (of an uncaught exception) is shown, but the internals of `raise` are not.
+    protocol = protocol or signal
+    condition = _prepare_signal_instance(condition, cause=cause, protocol=protocol, stacklevel=3)
+
     def accepts_arg(f):
         try:
             if arity_includes(f, 1):
@@ -157,6 +189,18 @@ def signal(condition, *, cause=None):
             return True  # just assume it
         return False
 
+    for handler in _find_handlers(type(condition)):
+        if accepts_arg(handler):
+            handler(condition)
+        else:
+            handler()
+
+    # When unhandled, return the condition instance.
+    # `error()` uses this return value; this allows us to provide a unified format for tracebacks.
+    return condition
+
+def _prepare_signal_instance(condition, *, cause, protocol, stacklevel):
+    """Canonize a condition, and populate its technical data."""
     # Consistency with behavior of exceptions in Python:
     #   Even if a class is raised, as in `raise StopIteration`, the `raise` statement
     #   converts it into an instance by instantiating with no args. So we need no
@@ -173,25 +217,17 @@ def signal(condition, *, cause=None):
                 return exc()  # instantiate with no args, like `raise` does
         except TypeError:  # "issubclass() arg 1 must be a class"
             pass
-        error(ControlError("Only exceptions and subclasses of Exception can {}; got {} with value {}.".format(err_reason, type(condition), repr(condition))))
+        error(ControlError(f"Only instances (derived too) and subclasses of BaseException can {err_reason}; got {type(condition)} with value {repr(condition)}."))
 
     condition = canonize(condition, "be signaled")
     cause = canonize(cause, "act as the cause of another signal")
     condition.__cause__ = cause
+    condition.__protocol__ = protocol
 
-    # Embed a stack trace in the signal, like Python does for raised exceptions.
-    # This only works on Python 3.7 and later, because we need to create a traceback object in pure Python code.
-    try:
-        # In the result, omit equip_with_traceback() and signal().
-        condition = equip_with_traceback(condition, stacklevel=2)
-    except NotImplementedError:  # pragma: no cover
-        pass  # well, we tried!
+    # Embed a stack trace in the signal, like Python does for raised exceptions. This API was added in Python 3.7.
+    condition = equip_with_traceback(condition, stacklevel=stacklevel)
 
-    for handler in _find_handlers(type(condition)):
-        if accepts_arg(handler):
-            handler(condition)
-        else:
-            handler()
+    return condition
 
 def invoke(name_or_restart, *args, **kwargs):
     """Invoke a restart currently in scope. Known as `INVOKE-RESTART` in Common Lisp.
@@ -225,11 +261,11 @@ def invoke(name_or_restart, *args, **kwargs):
     if isinstance(name_or_restart, str):
         restart = find_restart(name_or_restart)
         if not restart:
-            error(ControlError("No such restart: {}; available restarts: {}".format(repr(name_or_restart), available_restarts())))
+            error(ControlError(f"No such restart: {repr(name_or_restart)}; available restarts: {available_restarts()}"))
     elif isinstance(name_or_restart, BoundRestart):
         restart = name_or_restart
     else:
-        error(TypeError("Expected str or a return value of find_restart, got {} with value {}".format(type(name_or_restart), repr(name_or_restart))))
+        error(TypeError(f"Expected str or a return value of find_restart, got {type(name_or_restart)} with value {repr(name_or_restart)}"))
     # Found it - now we are guaranteed to unwind only up to the matching "with restarts".
     raise InvokeRestart(restart, *args, **kwargs)
 
@@ -241,23 +277,23 @@ Known as the `USE-VALUE` restart function in Common Lisp.
 A handler that just invokes the `use_value` restart is such a common use case
 that it is useful to have an abbreviation for it. This::
 
-    with handlers((OhNoes, lambda c: invoke("use_value", 42))):
+    with handlers((OhNoes, lambda: invoke("use_value", 42))):
         ...
 
 can be abbreviated to::
 
-    with handlers((OhNoes, lambda c: use_value(42))):
+    with handlers((OhNoes, lambda: use_value(42))):
         ...
 
-The `lambda c:` is still required, for consistency with Common Lisp, as well as
-to allow the user code to access the condition instance if needed.
-
-(A common use case is to embed, in the condition instance, the data needed
+A common use case is to embed, in the condition instance, the data needed
 for constructing the actual value to be sent to the `use_value` restart. In
 Seibel's log file parser example, when the parser sees a corrupt log entry,
 it embeds that data into the condition instance, and sends it to the handler,
 which then can in principle repair the log entry, and then invoke `use_value`
-with the repaired log entry.)
+with the repaired log entry. Then you can do something like::
+
+    with handlers((OhNoes, lambda c: use_value(produce_fixed_entry_from(c)))):
+        ...
 
 **Notes**:
 
@@ -270,26 +306,27 @@ The `use_value` function is essentially just shorthand::
 This pattern can be useful for defining similar shorthands for your own
 restarts.
 
-(Note that restarts are looked up by name, so a single module-level definition
+Note that restarts are looked up by name, so a single module-level definition
 of a shorthand for each uniquely named restart is enough. You can re-use the
 same shorthand for any restart that has the same name - just like there is
 just one `use_value` function, even though the `use_value` restart itself is
 defined separately at each `with restarts` site that provides it (since only
-each site itself knows how to "use a value").)
+each site itself knows how to "use a value").
 
 If you want a version for use cases where the condition instance argument is
 not needed, so you could in those cases omit the `lambda c:`, you can write
 that as::
 
-    use_constant = partial(invoker, "use_value")
-    with handlers((OhNoes, use_constant(42))):
+    make_use_constant = partial(invoker, "use_value")
+    use_42 = make_use_constant(42)
+    with handlers((OhNoes, use_42)):
         ...
 
 Note `invoker`, not `invoke`, and we are still left with a factory (since
 `invoker` itself is a factory and `partial` defers the call until it gets
 more arguments). You then call the factory function with your desired
 constant args/kwargs, to instantiate a handler that sends that specific
-set of args/kwargs.
+set of constant args/kwargs.
 """
 
 def invoker(restart_name, *args, **kwargs):
@@ -349,8 +386,21 @@ def invoker(restart_name, *args, **kwargs):
         with handlers((OhNoes, lambda c: use_value(42))):
             ...  # calling some code that may cerror(OhNoes("ouch"))
 
-    (The `use_value` function is convenient especially when the value being sent
-    is not a constant, but depends on data in the condition instance `c`.)
+    The `use_value` function is convenient especially when the value being sent
+    is not a constant, but depends on data in the condition instance `c`.
+    To do the same for your own restart, use this pattern (see `invoke`)::
+
+        frobnicate = partial(invoke, "frobnicate")
+        with handlers((OhNoes, frobnicate)):
+            ...
+
+    In this case, the `frobnicate` restart - if it accepts one positional
+    argument - will receive the condition instance. To send something else,
+    you can also do something like this::
+
+        frobnicate = partial(invoke, "frobnicate")
+        with handlers((OhNoes, lambda c: frobnicate(c.args[0] * 42))):
+            ...
 
     **Notes**
 
@@ -363,7 +413,7 @@ def invoker(restart_name, *args, **kwargs):
     """
     rename = namelambda(restart_name)
     the_invoker = rename(lambda c: invoke(restart_name, *args, **kwargs))
-    the_invoker.__doc__ = "Invoke the '{}' restart.".format(restart_name)
+    the_invoker.__doc__ = f"Invoke the '{restart_name}' restart."
     return the_invoker
 
 class _Stacked:  # boilerplate
@@ -445,7 +495,7 @@ class handlers(_Stacked):
         super().__init__(bindings)
         self.dq = _stacks.handlers
 
-class InvokeRestart(Exception):
+class InvokeRestart(BaseException):
     def __init__(self, restart, *args, **kwargs):  # e is the context
         self.restart, self.a, self.kw = restart, args, kwargs
         # message when uncaught
@@ -693,15 +743,7 @@ def error(condition, *, cause=None):
     keyword, it essentially performs a `error ... from ...`. The default
     `cause=None` performs a plain `error ...`.
     """
-    signal(condition, cause=cause)
-    # TODO: If we want to support the debugger at some point in the future,
-    # TODO: this is the appropriate point to ask the user what to do,
-    # TODO: before the call stack unwinds.
-    #
-    # TODO: Do we want to give one last chance to handle the ControlError?
-    # TODO: And do we want to raise ControlError, or the original condition?
-    condition.__cause__ = cause  # chain the causes, since we'll add a new one next.
-    raise ControlError("Unhandled error condition") from condition
+    _error(condition, cause=cause, protocol=error)
 
 def cerror(condition, *, cause=None):
     """Like `error`, but allow a handler to instruct the caller to ignore the error.
@@ -727,17 +769,31 @@ def cerror(condition, *, cause=None):
             def __init__(self, value):
                 self.value = value
 
-        with handlers=((OddNumberError, proceed)):
+        with handlers((OddNumberError, proceed)):
             out = []
             for x in range(10):
                 if x % 2 == 1:
                     cerror(OddNumberError(x))  # if unhandled, raises ControlError
                 out.append(x)
-        assert out == [0, 2, 4, 6, 8]
+        assert out == list(range(10))
 
     """
     with restarts(proceed=(lambda: None)):  # just for control, no return value
-        error(condition, cause=cause)
+        _error(condition, cause=cause, protocol=cerror)
+
+def _error(condition, *, cause, protocol):
+    # The return value is canonized to an instance (even if `condition` was an exception *type*),
+    # and importantly, it has a nice-looking traceback that points to this line here.
+    # If the signal goes unhandled, Python's exception system will want to show that traceback
+    # when our `ControlError` *exception* goes uncaught.
+    condition = signal(condition, cause=cause, protocol=protocol)
+    # TODO: If we want to support the debugger at some point in the future,
+    # TODO: this is the appropriate point to ask the user what to do,
+    # TODO: before the call stack unwinds.
+    #
+    # TODO: Do we want to give one last chance to handle the ControlError?
+    # TODO: And do we want to raise ControlError, or the original condition?
+    raise ControlError("Unhandled error condition") from condition
 
 def warn(condition, *, cause=None):
     """Like `signal`, but emit a warning if the condition is not handled.
@@ -787,7 +843,7 @@ def warn(condition, *, cause=None):
     """
     with restarts(muffle=(lambda: None)):  # just for control, no return value
         with restarts(_proceed=(lambda: None)):  # for internal use by unpythonic.test.fixtures
-            signal(condition, cause=cause)
+            signal(condition, cause=cause, protocol=warn)
         if isinstance(condition, Warning):
             warnings.warn(condition, stacklevel=2)  # 2 to ignore our lispy `warn` wrapper.
         else:
@@ -800,3 +856,107 @@ proceed.__doc__ = "Invoke the 'proceed' restart. Restart function for use with `
 
 muffle = invoker("muffle")
 muffle.__doc__ = "Invoke the 'muffle' restart. Restart function for use with `warn`."
+
+# Library to application signal type auto-conversion
+
+def _resignal_handler(mapping, condition):
+    """Remap a condition instance to another condition type.
+
+    `mapping`: dict-like, `{LibraryExc0: ApplicationExc0, ...}`
+
+        Each `LibraryExc` must be an exception type or a tuple of
+        exception types. It will be matched using `isinstance`.
+
+        Each `ApplicationExc` can be an exception type or an exception
+        instance. If an instance, then that exact instance is signaled
+        as the converted signal.
+
+    `condition`: the exception instance that was signaled, and is to
+                 be converted (if it matches an entry in `mapping`).
+                 When converted, it is automatically chained into
+                 an `ApplicationExc` signal.
+
+    Conversions in `mapping` are tried in the order specified; hence,
+    just like in `with handlers`, place more specific types first.
+
+    If no key in the mapping matches, this delegates to the next outer
+    signal handler.
+    """
+    for LibraryExc, ApplicationExc in mapping.items():
+        if isinstance(condition, LibraryExc):
+            # Resignal using the same error-handling protocol as the original signal
+            # (so that e.g. an `error(...)` resignals into an `error(...)` of the new type).
+            if not hasattr(condition, "__protocol__"):
+                error(f"Cannot resignal: protocol information missing in condition instance {condition}")
+            resignaler = condition.__protocol__
+            resignaler(ApplicationExc, cause=condition)
+    # cancel and delegate to the next outer handler
+
+def resignal_in(body, mapping):
+    """Remap condition types in an expression.
+
+    Like `unpythonic.excutil.reraise_in` (which see), but for conditions.
+
+    Usage::
+
+        resignal_in(body,
+                    {LibraryExc: ApplicationExc,
+                    ...})
+
+    Whenever `body` signals an `exc` for which it holds that
+    `isinstance(exc, LibraryExc)`, that signal will be transparently
+    chained into an `ApplicationExc` signal. The automatic conversion
+    is in effect for the dynamic extent of `body`.
+
+    ``body`` is a thunk (0-argument function).
+
+    ``mapping`` is dict-like, ``{input0: output0, ...}``, where each
+                 ``input``  is either an exception type,
+                            or a tuple of exception types.
+                            It will be matched using `isinstance`.
+                 ``output`` is an exception type or an exception
+                            instance. If an instance, then that exact
+                            instance is signaled as the converted
+                            signal.
+
+    Conversions are tried in the order specified; hence, just like in
+    `with handlers`, place more specific types first.
+
+    See also `resignal` for a block form.
+    """
+    with handlers((BaseException, partial(_resignal_handler, mapping))):
+        return body()
+
+@contextlib.contextmanager
+def resignal(mapping):
+    """Remap condition types. Context manager.
+
+    Like `unpythonic.excutil.reraise` (which see), but for conditions.
+
+    Usage::
+
+        with resignal({LibraryExc: ApplicationExc, ...}):
+            body0
+            ...
+
+    Whenever the body signals an `exc` for which it holds that
+    `isinstance(exc, LibraryExc)`, that signal will be transparently
+    chained into an `ApplicationExc` signal. The automatic conversion
+    is in effect for the dynamic extent of the `with` block.
+
+    ``mapping`` is dict-like, ``{input0: output0, ...}``, where each
+                 ``input``  is either an exception type,
+                            or a tuple of exception types.
+                            It will be matched using `isinstance`.
+                 ``output`` is an exception type or an exception
+                            instance. If an instance, then that exact
+                            instance is signaled as the converted
+                            signal.
+
+    Conversions are tried in the order specified; hence, just like in
+    `with handlers`, place more specific types first.
+
+    See also `resignal_in` for an expression form.
+    """
+    with handlers((BaseException, partial(_resignal_handler, mapping))):
+        yield
