@@ -19,16 +19,13 @@ Finally, we provide ready-made generators that yield some common sequences
 """
 
 __all__ = ["s", "imathify", "gmathify",
-           "m", "mg",  # old names, pre-0.14.3, will go away in 0.15.0
-           "almosteq",
            "sadd", "ssub", "sabs", "spos", "sneg", "sinvert", "smul", "spow",
            "struediv", "sfloordiv", "smod", "sdivmod",
            "sround", "strunc", "sfloor", "sceil",
            "slshift", "srshift", "sand", "sxor", "sor",
            "cauchyprod", "diagonal_reduce",
-           "fibonacci", "primes"]
+           "fibonacci", "triangular", "primes"]
 
-from warnings import warn
 from itertools import repeat, takewhile, count
 from functools import wraps
 from operator import (add as primitive_add, mul as primitive_mul,
@@ -45,13 +42,13 @@ from operator import (add as primitive_add, mul as primitive_mul,
 
 from .it import take, rev, window
 from .gmemo import imemoize, gmemoize
+from .numutil import almosteq
 
 class _NoSuchType:
     pass
 
 # stuff to support float, mpf and SymPy expressions transparently
 #
-from sys import float_info
 from math import log as math_log, copysign, trunc, floor, ceil
 try:
     from mpmath import mpf, almosteq as mpf_almosteq
@@ -60,6 +57,11 @@ except ImportError:  # pragma: no cover, optional at runtime, but installed at d
     mpf = _NoSuchType
     mpf_almosteq = None
 
+try:
+    import sympy
+except ImportError:  # pragma: no cover, optional at runtime, but installed at development time.
+    sympy = None
+
 def _numsign(x):
     """The sign function, for numeric inputs."""
     if x == 0:
@@ -98,44 +100,6 @@ except ImportError:  # pragma: no cover, optional at runtime, but installed at d
     sign = _numsign
     _symExpr = _NoSuchType
 
-# TODO: Overhaul `almosteq` in v0.15.0, should work like mpf for consistency.
-# TODO: Also move it to `unpythonic.misc`, where `ulp` already is. Or make a `numutil`.
-def almosteq(a, b, tol=1e-8):
-    """Almost-equality that supports several formats.
-
-    The tolerance ``tol`` is used for the builtin ``float`` and ``mpmath.mpf``.
-
-    For ``mpmath.mpf``, we just delegate to ``mpmath.almosteq``, with the given
-    ``tol``. For ``float``, we use the strategy suggested in:
-
-        https://floating-point-gui.de/errors/comparison/
-
-    Anything else, for example SymPy expressions, strings, and containers
-    (regardless of content), is tested for exact equality.
-
-    **CAUTION**: Although placed in ``unpythonic.mathseq``, this function
-    **does not** support iterables; rather, it is a low-level tool that is
-    exposed in the public API in the hope it may be useful elsewhere.
-    """
-    if a == b:  # infinities and such, plus any non-float type
-        return True
-
-    if isinstance(a, mpf) and isinstance(b, mpf):
-        return mpf_almosteq(a, b, tol)
-    # compare as native float if only one is an mpf
-    elif isinstance(a, mpf) and isinstance(b, (float, int)):
-        a = float(a)
-    elif isinstance(a, (float, int)) and isinstance(b, mpf):
-        b = float(b)
-
-    if not all(isinstance(x, (float, int)) for x in (a, b)):
-        return False  # non-float type, already determined that a != b
-    min_normal = float_info.min
-    max_float = float_info.max
-    d = abs(a - b)
-    if a == 0 or b == 0 or d < min_normal:
-        return d < tol * min_normal
-    return d / min(abs(a) + abs(b), max_float) < tol
 
 def s(*spec):
     """Create a lazy mathematical sequence.
@@ -306,6 +270,8 @@ def s(*spec):
 
     def is_almost_int(x):
         try:
+            if sympy and isinstance(x, sympy.Expr):
+                x = sympy.N(x)
             return almosteq(float(round(x)), x)
         except TypeError:  # likely a SymPy expression that didn't simplify to a number
             return False
@@ -351,9 +317,9 @@ def s(*spec):
                     p = (p1 + p2) / 2
                     return ("power", a0, p)
             # Most unrecognized sequences trigger this case.
-            raise SyntaxError("Specification did not match any supported formula: '{}'".format(origspec))
+            raise SyntaxError(f"Specification did not match any supported formula: '{origspec}'")
         else:  # more elements are optional but must be consistent
-            data = [analyze(*triplet) for triplet in window(iterable=spec, n=3)]
+            data = [analyze(*triplet) for triplet in window(3, spec)]
             seqtypes, x0s, ks = zip(*data)
             def isconst(xs):
                 first, *rest = xs
@@ -361,7 +327,7 @@ def s(*spec):
             if not isconst(seqtypes) or not isconst(ks):
                 # This case is only triggered if all triplets specify some
                 # recognized sequence, but the specifications don't agree.
-                raise SyntaxError("Inconsistent specification '{}'".format(origspec))
+                raise SyntaxError(f"Inconsistent specification '{origspec}'")
             return data[0]
 
     # final term handler for finite sequences - compute how many terms we should generate in total
@@ -411,7 +377,7 @@ def s(*spec):
         *spec, last = spec
         if last is Ellipsis:
             if not spec:
-                raise SyntaxError("Expected s(a0, a1, ...), s(a0, a1, ..., an), s([*repeats], ...), or s(*initials, [*repeats], ...); got '{}'".format(origspec))
+                raise SyntaxError(f"Expected s(a0, a1, ...), s(a0, a1, ..., an), s([*repeats], ...), or s(*initials, [*repeats], ...); got '{origspec}'")
             assert spec  # not empty
             # v0.14.3+: cyclic infinite sequences
             if iscyclic(spec):
@@ -424,14 +390,14 @@ def s(*spec):
         else:
             *spec, dots = spec
             if not (dots is Ellipsis and spec):
-                raise SyntaxError("Expected s(a0, a1, ...) or s(a0, a1, ..., an), s([*repeats], ...), or s(*initials, [*repeats], ...); got '{}'".format(origspec))
+                raise SyntaxError(f"Expected s(a0, a1, ...) or s(a0, a1, ..., an), s([*repeats], ...), or s(*initials, [*repeats], ...); got '{origspec}'")
             assert spec  # not empty
             desc = analyze(*spec)
             n = nofterms(desc, last)
             if n is False:
-                raise SyntaxError("The final element, if present, must belong to the specified sequence; got '{}'".format(origspec))
+                raise SyntaxError(f"The final element, if present, must belong to the specified sequence; got '{origspec}'")
             elif n is infty:
-                raise SyntaxError("The length of a constant sequence cannot be determined from a final element; got '{}'".format(origspec))
+                raise SyntaxError(f"The length of a constant sequence cannot be determined from a final element; got '{origspec}'")
             seqtype, x0, k = desc
 
     # generate the sequence
@@ -454,7 +420,7 @@ def s(*spec):
         return imathify(arith() if n is infty else take(n, arith()))
     elif seqtype == "geom":
         if isinstance(k, _symExpr) or abs(k) >= 1:
-            def geoimathify():
+            def geom():
                 j = 0
                 while True:
                     yield x0 * (k**j)
@@ -466,12 +432,12 @@ def s(*spec):
             # Note that 1/(1/3) --> 3.0 even for floats, so we don't actually
             # need to modify the detection algorithm to account for this.
             kinv = 1 / k
-            def geoimathify():
+            def geom():
                 j = 0
                 while True:
                     yield x0 / (kinv**j)
                     j += 1
-        return imathify(geoimathify() if n is infty else take(n, geoimathify()))
+        return imathify(geom() if n is infty else take(n, geom()))
     else:  # seqtype == "power":
         if isinstance(k, _symExpr) or abs(k) >= 1:
             def power():
@@ -634,14 +600,6 @@ class imathify:
     def __gt__(self, other):
         return sgt(self, other)
 
-class m(imathify):  # pragma: no cover
-    """Alias for `imathify`, for backward compatibility.
-
-    Will be removed in 0.15.0."""
-    def __init__(self, iterable):
-        warn("`m` has been renamed `imathify`, which is more descriptive; this alias will be removed in 0.15.0.", FutureWarning)
-        super().__init__(iterable)
-
 def gmathify(gfunc):
     """Decorator: make gfunc imathify() the returned generator instances.
 
@@ -659,14 +617,6 @@ def gmathify(gfunc):
         return imathify(gfunc(*args, **kwargs))
     return mathify
 
-def mg(gfunc):  # pragma: no cover
-    """Alias for `gmathify`, for backward compatibility.
-
-    Will be removed in 0.15.0.
-    """
-    warn("`mg` has been renamed `gmathify`, which is more descriptive; this alias will be removed in 0.15.0.", FutureWarning)
-    return gmathify(gfunc)
-
 # -----------------------------------------------------------------------------
 # We expose the full set of "imathify" operators also as functions à la the ``operator`` module.
 # Prefix "s", short for "mathematical Sequence".
@@ -911,10 +861,11 @@ def diagonal_reduce(a, b, *, combine, reduce, require="any"):
     is not formed, because the terms ``a[0]*b[2]`` and ``a[2]*b[0]`` (that would
     contribute to it in the infinite case) cannot be formed from length-2 inputs.
     """
+    # TODO: Python 3.8+: test for the appropriate `typing` Protocol instead?
     if not all(hasattr(x, "__iter__") for x in (a, b)):
-        raise TypeError("Expected two iterables, got {}, {}".format(type(a), type(b)))
+        raise TypeError(f"Expected two iterables, got {type(a)}, {type(b)}")
     if require not in ("all", "any"):
-        raise ValueError("require must be 'all' or 'any'; got '{}'".format(require))
+        raise ValueError(f"require must be 'all' or 'any'; got '{require}'")
     ga = imemoize(a)
     gb = imemoize(b)
     def diagonal():
@@ -945,6 +896,28 @@ def fibonacci():
             a, b = b, a + b
     return imathify(fibos())
 
+def triangular():
+    """Return the triangular numbers 1, 3, 6, 10, ... as a lazy sequence.
+
+    Etymology::
+
+            x
+           x x
+          x x x
+         x x x x
+        ...
+    """
+    # We could just use Gauss's result  n * (n + 1) / 2  (which can be proved by induction),
+    # but this algorithm is trivially correct.
+    def _triangular():
+        s = 1  # running total
+        r = 2  # places in the next row of the triangle
+        while True:
+            yield s
+            s += r
+            r += 1
+    return imathify(_triangular())
+
 # See test_gmemo.py for history. This is an FP-ized sieve of Eratosthenes.
 #
 # This version wins in speed for moderate n (1e5) on typical architectures where
@@ -983,7 +956,7 @@ def primes(optimize="speed"):
     general ``gmemoize`` machinery in the inner loop).
     """
     if optimize not in ("memory", "speed"):
-        raise ValueError("optimize must be 'memory' or 'speed'; got '{}'".format(optimize))
+        raise ValueError(f"optimize must be 'memory' or 'speed'; got '{optimize}'")
     if optimize == "speed":
         return imathify(_fastprimes())
     else:  # optimize == "memory":
