@@ -80,6 +80,7 @@ from ast import (Name, Tuple, Lambda, FunctionDef, AsyncFunctionDef, ClassDef,
                  Import, ImportFrom, Try, ListComp, SetComp, GeneratorExp,
                  DictComp, Store, Del, Global, Nonlocal)
 
+from mcpyrate.astcompat import TryStar, MatchStar, MatchMapping, MatchClass, MatchAs
 from mcpyrate.core import Done
 from mcpyrate.walkers import ASTTransformer, ASTVisitor
 
@@ -289,8 +290,8 @@ def get_names_in_store_context(tree):
 
     This includes:
 
-        - Any ``Name`` in store context (such as on the LHS of an `Assign`
-          or `NamedExpr` node)
+        - Any ``Name`` in store context (such as on the LHS of an `Assign`,
+          `NamedExpr` (Python 3.8+), `TypeAlias` (Python 3.12+))
 
         - The name of ``FunctionDef``, ``AsyncFunctionDef`` or``ClassDef``
 
@@ -300,8 +301,12 @@ def get_names_in_store_context(tree):
 
         - The exception name of any ``except`` handlers
 
+        - The exception name of any ``except*`` handlers (Python 3.11+)
+
         - The names in the as-part of ``With``
 
+        - The names bound in `match`/`case` patterns (Python 3.10+)
+
     Duplicates may be returned; use ``set(...)`` or ``list(uniqify(...))``
     on the output to remove them.
 
@@ -311,6 +316,12 @@ def get_names_in_store_context(tree):
     by ``get_lexical_variables`` for the nearest lexically surrounding parent
     tree that represents a scope.
     """
+    class MatchCapturesCollector(ASTVisitor):  # Python 3.10+: `match`/`case`
+        def examine(self, tree):
+            if type(tree) is Name:
+                self.collect(tree.id)
+            self.generic_visit(tree)
+
     class StoreNamesCollector(ASTVisitor):
         # def _collect_name_or_list(self, t):
         #     if type(t) is Name:
@@ -332,7 +343,7 @@ def get_names_in_store_context(tree):
             elif type(tree) in (Import, ImportFrom):
                 for x in tree.names:
                     self.collect(x.asname if x.asname is not None else x.name)
-            elif type(tree) is Try:
+            elif type(tree) in (Try, TryStar):  # Python 3.11+: `try`/`except*`
                 # https://docs.python.org/3/reference/compound_stmts.html#the-try-statement
                 #
                 # TODO: The `err` in  `except SomeException as err` is only bound within the `except` block,
@@ -344,6 +355,31 @@ def get_names_in_store_context(tree):
                 # TODO: `try`, even inside the `except` blocks, will be bound in the whole parent scope.
                 for h in tree.handlers:
                     self.collect(h.name)
+            # Python 3.10+: `match`/`case` uses names in `Load` context to denote captures.
+            # Also there are some bare strings, and sometimes `None` actually means "_" (but doesn't capture).
+            # So we special-case all of this.
+            elif type(tree) in (MatchAs, MatchStar):  # a `MatchSequence` also consists of these
+                if tree.name is not None:
+                    self.collect(tree.name)
+            elif type(tree) is MatchMapping:
+                mcc = MatchCapturesCollector(tree.patterns)
+                mcc.visit()
+                for name in mcc.collected:
+                    self.collect(name)
+                if tree.rest is not None:  # `rest` is a capture if present
+                    self.collect(tree.rest)
+            elif type(tree) is MatchClass:
+                mcc = MatchCapturesCollector(tree.patterns)
+                mcc.visit()
+                for name in mcc.collected:
+                    self.collect(name)
+                mcc = MatchCapturesCollector(tree.kwd_patterns)
+                mcc.visit()
+                for name in mcc.collected:
+                    self.collect(name)
+
+            # Python 3.12+: `TypeAlias` uses a name in `Store` context on its LHS so it needs no special handling here.
+
             # Same note as for for loops.
             # elif type(tree) in (With, AsyncWith):
             #     for item in tree.items:
@@ -369,7 +405,7 @@ def get_names_in_del_context(tree):
         def examine(self, tree):
             # We want to detect things like "del x":
             #     Delete(targets=[Name(id='x', ctx=Del()),])
-            # We don't currently care about "del myobj.x" or "del mydict['x']" (these examples in Python 3.6):
+            # We don't currently care about "del myobj.x" or "del mydict['x']" (these old examples in Python 3.6):
             #     Delete(targets=[Attribute(value=Name(id='myobj', ctx=Load()), attr='x', ctx=Del()),])
             #     Delete(targets=[Subscript(value=Name(id='mydict', ctx=Load()), slice=Index(value=Str(s='x')), ctx=Del()),])
             if type(tree) is Name and hasattr(tree, "ctx") and type(tree.ctx) is Del:
